/*
 * Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may not use this file except in compliance with
 * the License. A copy of the License is located at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 * CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions
 * and limitations under the License.
 */

import { isPlainObject } from 'is-plain-object';
import deepExtend from '../utils/deepExtend.js';
import GroupMessages from '../utils/groupMessages.js';

const MISSING_TRANSFORM_ERRORS = GroupMessages.GROUP.MissingRegisterTransformErrors;

/**
 * Takes a platform platformConfig object and returns a new one
 * that has filters, transforms, formats, and actions
 * mapped properly.
 * @private
 * @param {Object} platformConfig
 * @param {Object} dictionary
 * @param {Object} platformName (only used for error messaging)
 * @returns {Object}
 */
export default function transformConfig(platformConfig, dictionary, platformName) {
  const to_ret = { ...platformConfig }; // structuredClone not suitable due to config being able to contain Function() etc.
  to_ret.log = platformConfig.log ?? dictionary.log;

  // The platform can define either a transformGroup or an array
  // of transforms. If given a transformGroup that doesn't exist,
  // it will throw an error to make the user aware that the transformGroup doesn't
  // exist. A valid case is if the user defines neither, no transforms will be
  // applied.
  let transforms = [];
  if (to_ret.transforms) {
    transforms = to_ret.transforms;
  } else if (to_ret.transformGroup) {
    if (dictionary.transformGroup[to_ret.transformGroup]) {
      transforms = dictionary.transformGroup[to_ret.transformGroup];
    } else {
      let err = `
Unknown transformGroup "${to_ret.transformGroup}" found in platform "${platformName}":
"${to_ret.transformGroup}" does not match the name of a registered transformGroup.
`;
      throw new Error(err);
    }
  }

  // Transforms are an array of strings that map to functions on
  // the StyleDictionary module. We need to map the strings to
  // the actual functions.
  to_ret.transforms = transforms.map(function (name) {
    if (!dictionary.transform[name]) {
      GroupMessages.add(MISSING_TRANSFORM_ERRORS, `"${name}"`);
    }
    return dictionary.transform[name];
  });

  let missingTransformCount = GroupMessages.count(MISSING_TRANSFORM_ERRORS);
  if (missingTransformCount > 0) {
    const transform_warnings = GroupMessages.flush(MISSING_TRANSFORM_ERRORS).join(', ');
    let err;

    if (missingTransformCount == 1) {
      err = `
Unknown transform ${transform_warnings} found in platform "${platformName}":
${transform_warnings} does not match the name of a registered transform.
`;
    } else {
      err = `
Unknown transforms ${transform_warnings} found in platform "${platformName}":
None of ${transform_warnings} match the name of a registered transform.
`;
    }

    throw new Error(err);
  }

  // Apply registered fileHeaders onto the platform options
  if (platformConfig.options && platformConfig.options.fileHeader) {
    const fileHeader = platformConfig.options.fileHeader;
    if (typeof fileHeader === 'string') {
      if (dictionary.fileHeader[fileHeader]) {
        to_ret.options.fileHeader = dictionary.fileHeader[fileHeader];
      } else {
        throw new Error(`Can't find fileHeader: ${fileHeader}`);
      }
    } else if (typeof fileHeader !== 'function') {
      throw new Error(`fileHeader must be a string or a function`);
    } else {
      to_ret.options.fileHeader = fileHeader;
    }
  }

  to_ret.files = (platformConfig.files || []).map(function (file) {
    const ext = { options: {} };
    if (file.options && file.options.fileHeader) {
      const fileHeader = file.options.fileHeader;
      if (typeof fileHeader === 'string') {
        if (dictionary.fileHeader[fileHeader]) {
          ext.options.fileHeader = dictionary.fileHeader[fileHeader];
        } else {
          throw new Error(`Can't find fileHeader: ${fileHeader}`);
        }
      } else if (typeof fileHeader !== 'function') {
        throw new Error(`fileHeader must be a string or a function`);
      } else {
        ext.options.fileHeader = fileHeader;
      }
    }

    if (file.filter) {
      if (typeof file.filter === 'string') {
        if (dictionary.filter[file.filter]) {
          ext.filter = dictionary.filter[file.filter];
        } else {
          throw new Error("Can't find filter: " + file.filter);
        }
      } else if (typeof file.filter === 'object') {
        // Recursively go over the object keys of filter object and
        // return a filter Function that filters tokens
        // by the specified object keys.
        const matchFn = function (inputObj, testObj) {
          if (isPlainObject(testObj)) {
            return Object.keys(testObj).every((key) => matchFn(inputObj[key], testObj[key]));
          } else {
            return inputObj == testObj;
          }
        };
        const matches = function (matchObj) {
          let cloneObj = { ...matchObj }; // shallow clone, structuredClone not suitable because obj can contain "Function()"
          let matchesFn = (inputObj) => matchFn(inputObj, cloneObj);
          return matchesFn;
        };
        ext.filter = matches(file.filter);
      } else if (typeof file.filter === 'function') {
        ext.filter = file.filter;
      } else {
        throw new Error('Filter format not valid: ' + typeof file.filter);
      }
    }

    if (file.format) {
      if (dictionary.format[file.format]) {
        ext.format = dictionary.format[file.format];
      } else {
        throw new Error("Can't find format: " + file.format);
      }
    } else {
      throw new Error('Please supply a format for file: ' + JSON.stringify(file));
    }
    return deepExtend([{}, file, ext]);
  });

  to_ret.actions = (platformConfig.actions || []).map(function (action) {
    if (typeof dictionary.action[action].undo !== 'function') {
      console.warn(action + ' action does not have a clean function!');
    }
    return dictionary.action[action];
  });

  return to_ret;
}
